import typing as tp
from collections import defaultdict

from airflow.models.baseoperator import BaseOperator
from airflow.utils.task_group import TaskGroup

if tp.TYPE_CHECKING:
    from dbt_af.builder.dag_components import DagComponent


class DagDelayedDependency:
    """
    Delayed fixed pair of two tasks in Airflow DAG that have dependency between.
    It works because in Airflow it's possible to set dependency between tasks at any time
    after tasks and Dag are created.
    """

    def __init__(self, upstream: BaseOperator, downstream: BaseOperator):
        self.upstream = upstream
        self.downstream = downstream

    def resolve(self):
        self.upstream.set_downstream(self.downstream)


class DagDelayedDependencyStream:
    """
    Helper class to create delayed dependency between two tasks in Airflow DAG.
    It represents a dependency between two tasks in Airflow DAG using >> and << operators.
    """

    def __init__(self, stream: BaseOperator, registry):
        self.stream = stream
        self.registry = registry

    def __rshift__(self, other: 'DagDelayedDependencyStream'):
        if not isinstance(other, DagDelayedDependencyStream):
            raise TypeError(f'Cannot create dependency between {self} and {other}')
        if self.stream is None or other.stream is None:
            return self

        self.registry.register(self.stream, other.stream)
        return self

    def __lshift__(self, other: 'DagDelayedDependencyStream'):
        if not isinstance(other, DagDelayedDependencyStream):
            raise TypeError(f'Cannot create dependency between {self} and {other}')
        if self.stream is None or other.stream is None:
            return self

        self.registry.register(other.stream, self.stream)
        return self


class DagDelayedDependencyRegistry:
    """
    Registry of delayed dependencies between tasks in Airflow DAG. It's used to create delayed dependencies and
    resolve them at the end of the DagComponent creation.
    Supports context manager interface to resolve dependencies at the exit of the context.
    Correctly resolves dependencies between TaskGroup >> TaskGroup and TaskGroup >> Task

    Example:
        with DagDelayedDependencyRegistry() as delayed_deps:
            delayed_deps(group1) >> delayed_deps(task2)
            delayed_deps(task1) >> delayed_deps(task2)
            delayed_deps(group1) >> delayed_deps(group2)

    It's the same as (at the end of the function):
        group1 >> group2  # this dependency will be resolved first
        group1 >> task2
        task1 >> task2
    """

    def __init__(self):
        self._registry: tp.List[DagDelayedDependency] = []

    def __call__(self, task: tp.Union[BaseOperator, TaskGroup]) -> DagDelayedDependencyStream:
        return DagDelayedDependencyStream(task, registry=self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.resolve_dependencies()

    def __iter__(self):
        return iter(self._registry)

    def register(self, upstream: BaseOperator, downstream: BaseOperator):
        self._registry.append(DagDelayedDependency(upstream, downstream))

    def _sort_dependencies(self):
        # HACK: https://github.com/apache/airflow/issues/16764#issuecomment-1015058864
        # we need to resolve TaskGroup >> TaskGroup dependencies first and then all other dependencies
        # it's necessary to avoid bug with inconsistent state of TaskGroup and dependencies between them.
        self._registry.sort(
            key=lambda d: (isinstance(d.upstream, TaskGroup)) + (isinstance(d.downstream, TaskGroup)),
            reverse=True,
        )

    def resolve_dependencies(self):
        self._sort_dependencies()
        for dep in self._registry:
            dep.resolve()


class RegistryDomainDependencies:
    def __init__(self):
        self._registry: tp.Dict['DagComponent', tp.List[BaseOperator]] = defaultdict(list)
        self.task_group: tp.Optional[TaskGroup] = None

    def __repr__(self):
        return f'{self.__class__.__name__}({self._registry})'

    def is_registered(self, component: 'DagComponent') -> bool:
        return component in self._registry

    def add_dependency(self, component: 'DagComponent', wait_task: BaseOperator) -> None:
        self._registry[component].append(wait_task)

    def get_dependency_wait_task(self, component: 'DagComponent') -> tp.List[BaseOperator]:
        return self._registry[component]



import base64
import json
import logging
import os
import shlex
from typing import TYPE_CHECKING

import kubernetes.client.models as k8s
from airflow.configuration import conf

# this is a workaround to be compatible with old versions of Airflow and cncf-kubernetes provider
try:
    from airflow.providers.cncf.kubernetes.operators.kubernetes_pod import KubernetesPodOperator
except (ModuleNotFoundError, ImportError):
    from airflow.providers.cncf.kubernetes.operators.pod import KubernetesPodOperator

from dbt_af.common.constants import AZ_MI_BINDING_LABEL_NAME
from dbt_af.conf import Config
from dbt_af.parser.dbt_profiles import KubernetesTarget

if TYPE_CHECKING:
    from airflow.utils.context import Context

PYTHON_SCRIPT_ENV = '__PYTHON_SCRIPT'
PYTHON_SCRIPT_FILE = 'script.py'


class DbtKubernetesPodOperator(KubernetesPodOperator):
    def __init__(
        self,
        dbt_model_name: str,
        dbt_model_path: str,
        target_details: KubernetesTarget,
        dbt_af_config: Config,
        env: dict[str, str] | None = None,
        *args,
        **kwargs,
    ):
        self.dbt_model_name = dbt_model_name
        self.dbt_model_path = dbt_model_path
        self.target_details = target_details
        self.dbt_af_config = dbt_af_config

        super().__init__(
            *args,
            max_active_tis_per_dag=1,
            cmds=['bash', '-c', ' && '.join(self._prepare_docker_cmds())],
            in_cluster=True,
            get_logs=True,
            log_events_on_failure=True,
            do_xcom_push=True,
            startup_timeout_seconds=1200,
            env_vars=[k8s.V1EnvVar(name=k, value=v) for k, v in (env or {}).items()],
            namespace=conf.get('kubernetes_executor', 'NAMESPACE'),
            **dbt_af_config.retries_config.k8s_task_retry_policy.as_dict(),
            **kwargs,
        )

    def _find_model_config_by_name(self):
        with open(self.dbt_af_config.dbt_project.dbt_project_path / 'target/manifest.json', 'r') as fn:
            manifest = json.load(fn)

        model_config = {}
        for node_info in manifest['nodes'].values():
            if node_info['name'] == self.dbt_model_name:
                model_config = node_info['config']
                break
        logging.info(f'Fetched model config for {self.dbt_model_name}: {model_config}')

        return model_config

    @staticmethod
    def _prepare_docker_cmds():
        _file_name = shlex.quote(PYTHON_SCRIPT_FILE)
        make_xcom_dir_cmd = 'mkdir -p /airflow/xcom'
        write_source_cmd = (
            f'python -c "import os;'
            rf'x = os.environ[\"{PYTHON_SCRIPT_ENV}\"];'
            rf'f = open(\"{_file_name}\", \"w\"); f.write(x); f.close()"'
        )
        exec_script_cmd = f'python {_file_name}'
        return [
            make_xcom_dir_cmd,
            write_source_cmd,
            exec_script_cmd,
        ]

    def execute(self, context: 'Context'):
        model_config = self._find_model_config_by_name()
        model_config_b64 = base64.b64encode(json.dumps(model_config).encode()).decode()

        self.env_vars.append(k8s.V1EnvVar(name='MODEL_CONFIG_B64', value=model_config_b64))
        self.env_vars.append(k8s.V1EnvVar(name='START_DTTM', value=context['data_interval_start'].isoformat()))
        self.env_vars.append(k8s.V1EnvVar(name='END_DTTM', value=context['data_interval_end'].isoformat()))
        self.env_vars.append(k8s.V1EnvVar(name='AIRFLOW_UNIQUE_NAME', value=os.getenv('AIRFLOW_UNIQUE_NAME')))
        self.env_vars.append(k8s.V1EnvVar(name='DAG_RUN_CONF', value=json.dumps(context['dag_run'].conf or {})))

        with open(self.dbt_af_config.dbt_project.dbt_models_path / self.dbt_model_path, 'r') as f:
            raw_node_source = f.read()
        self.env_vars.append(k8s.V1EnvVar(name=PYTHON_SCRIPT_ENV, value=raw_node_source))

        self.env_vars.extend(
            [
                k8s.V1EnvVar(
                    name='AIRFLOW__SECRETS__BACKEND_KWARGS', value=os.getenv('AIRFLOW__SECRETS__BACKEND_KWARGS', '{}')
                ),
                k8s.V1EnvVar(name='AIRFLOW__SECRETS__BACKEND', value=os.getenv('AIRFLOW__SECRETS__BACKEND', '')),
            ]
        )

        if self.dbt_af_config.k8s.airflow_identity_binding_selector:
            # TODO: migrate to workload identity (https://learn.microsoft.com/en-us/azure/aks/workload-identity-migrate-from-pod-identity#deploy-the-workload-with-migration-sidecar)
            self.labels = {AZ_MI_BINDING_LABEL_NAME: self.dbt_af_config.k8s.airflow_identity_binding_selector}

        self.image = self.target_details.image_name
        self.tolerations = [
            k8s.V1Toleration(key=toleration.key, operator=toleration.operator)
            for toleration in self.target_details.tolerations
        ]
        self.container_resources = k8s.V1ResourceRequirements(
            limits={
                'cpu': self.target_details.pod_cpu_guarantee,
                'memory': self.target_details.pod_memory_guarantee,
            },
        )
        self.node_selector = {self.target_details.node_pool_selector_name: self.target_details.node_pool}

        super().execute(context)



import json
from pathlib import Path

import pytest
import yaml

from dbt_af.builder.dbt_af_builder import DbtAfGraph
from dbt_af.conf import Config, DbtDefaultTargetsConfig, DbtProjectConfig


def pytest_addoption(parser):
    parser.addoption('--manifest_path', action='store', default='.', help='Path to manifest.json')
    parser.addoption('--profiles_path', action='store', default='.', help='Path to profiles.yml')
    parser.addoption('--dbt_project_path', action='store', default='.', help='Path to dbt_project.yml')
    parser.addoption('--etl_name', action='store', default=None, help='List of ETLs to test')
    parser.addoption('--target', action='store', default='dev', help='Default target to use')


@pytest.fixture(scope='session')
def get_manifest_path(pytestconfig):
    return pytestconfig.getoption('manifest_path')


@pytest.fixture(scope='session')
def get_profiles_path(pytestconfig):
    return pytestconfig.getoption('profiles_path')


@pytest.fixture(scope='session')
def get_dbt_project_path(pytestconfig):
    return pytestconfig.getoption('dbt_project_path')


@pytest.fixture(scope='session')
def get_etl_name(pytestconfig):
    return pytestconfig.getoption('etl_name')


@pytest.fixture(scope='session')
def target(pytestconfig):
    return pytestconfig.getoption('target')


@pytest.fixture
def config(target):
    return Config(
        dbt_project=DbtProjectConfig(
            dbt_project_name='dwh',
            dbt_models_path='.',
            dbt_project_path='.',
            dbt_profiles_path='.',
            dbt_target_path='.',
            dbt_log_path='.',
            dbt_schema='.',
        ),
        dbt_default_targets=DbtDefaultTargetsConfig(default_target=target),
    )


@pytest.fixture
def manifest(get_manifest_path) -> dict:
    manifest_path = (
        get_manifest_path if get_manifest_path.endswith('manifest.json') else Path(get_manifest_path) / 'manifest.json'
    )
    with open(manifest_path, 'r') as f:
        return json.load(f)


@pytest.fixture
def profiles(get_profiles_path) -> dict:
    profiles_path = (
        get_profiles_path if get_profiles_path.endswith('profiles.yml') else Path(get_profiles_path) / 'profiles.yml'
    )
    with open(profiles_path, 'r') as f:
        return yaml.safe_load(f)


@pytest.fixture
def dbt_project(get_dbt_project_path):
    dbt_project_path = (
        get_dbt_project_path
        if get_dbt_project_path.endswith('dbt_project.yml')
        else Path(get_dbt_project_path) / 'dbt_project.yml'
    )
    with open(dbt_project_path, 'r') as f:
        return yaml.safe_load(f)


@pytest.fixture
def project_profile_name(dbt_project):
    return dbt_project['profile']


@pytest.fixture
def dbt_af_graph(manifest, profiles, project_profile_name, config):
    return DbtAfGraph.from_manifest(manifest, profiles, project_profile_name, etl_service_name='', config=config)



from dbt_af.parser.dbt_profiles import Profile


def test_kubernetes_profiles_have_correct_structure(profiles):
    for profile_name, profile in profiles.items():
        if profile_name == 'config':
            continue
        profile = Profile(**profile)
        for target_name, target in profile.outputs.items():
            if target.target_type == 'kubernetes':
                assert target.node_pool_selector_name is not None
                assert target.node_pool is not None
                assert target.image_name is not None
                assert target.pod_cpu_guarantee is not None
                assert target.pod_memory_guarantee is not None
                assert target.tolerations is not None



import logging
import os
from enum import Enum
from pathlib import Path
from typing import Optional

from airflow.models import Variable
from airflow.utils.context import Context
from cachetools import TTLCache, cached

from dbt_af.conf import Config


class TestTag(Enum):
    """DBT tag that specifies a test scheduling strategy in an Airflow DAG."""

    small = '@small'
    medium = '@medium'
    large = '@large'


def get_variable(var_name):
    if var_name in os.environ:
        return os.environ.get(var_name)

    return Variable.get(var_name, f'{{{{ var.value.{var_name} }}}}')


@cached(cache=TTLCache(maxsize=1024, ttl=60 * 5))
def init_environment(config: Config):
    dbt_env = {
        'DBT_TARGET_PATH': str(config.dbt_project.dbt_target_path),
        'DBT_LOG_PATH': str(config.dbt_project.dbt_log_path),
        'DBT_SCHEMA': config.dbt_project.dbt_schema,
        'DBT_PROJECT_DIR': str(config.dbt_project.dbt_project_path),
        'DBT_DEPS_DIR': os.path.join(config.dbt_project.dbt_project_path, 'dbt_packages'),
        'DBT_PROFILES_DIR': str(config.dbt_project.dbt_profiles_path),
        **config.dbt_project.additional_dbt_env,
    }

    return dbt_env


def find_latest_log_file(context: 'Context', log_dir: Path) -> Optional[str]:
    log_pattern = f'dag_id={context["dag"].dag_id}/run_id={context["run_id"]}/task_id={context["task"].task_id}/*.log'
    try:
        return str(max(log_dir.glob(log_pattern)))
    except TypeError:
        logging.error(f'No log file found for dag_id={context["dag"].dag_id}, task_id={context["task"].task_id}')
    except Exception as ex:
        logging.error(f'Something went wrong while searching log file, {ex}')

    return None



import unittest
from collections import defaultdict

from dbt_af.builder.dbt_af_builder import DbtAfGraph
from dbt_af.dags import dbt_main_dags
from dbt_af.parser.dbt_node_model import DbtNode


def _model_safe_name(node: DbtNode) -> str:
    return node.fqn[-1].replace('.', '__')


def _model_schedule_safe_name(schedule_name: str) -> str:
    return schedule_name.replace('@', '')


def check_if_all_dbt_models_are_in_dbt_af_graph(manifest, profiles, project_profile_name, config, etl=None):
    dbt_graph = DbtAfGraph.from_manifest(
        manifest=manifest,
        profiles=profiles,
        project_profile_name=project_profile_name,
        etl_service_name=etl,
        config=config,
    )
    dags = dbt_main_dags(dbt_graph)

    raw_dbt_nodes_by_domain_schedule = defaultdict(lambda: defaultdict(set))
    for raw_node in manifest['nodes'].values():
        node = DbtNode(**raw_node)
        if not node.is_at_etl_service(etl) or not node.is_model():
            continue

        raw_dbt_nodes_by_domain_schedule[node.domain][_model_schedule_safe_name(node.config.schedule.name)].add(
            _model_safe_name(node)
        )

    for dag_name, dag in dags.items():
        if (
            'dbt' not in dag.tags
            or 'backfill' in dag.tags
            or 'frontier' not in dag.tags
            or 'dbt_large_tests' in dag.tags
        ):
            continue

        domain_name, schedule, *_ = dag_name.split('__')
        task_ids = [
            task.task_id.split('.')[-1]
            for task in dag.tasks
            if task.operator_name not in ('DbtTest', 'EmptyOperator', 'DbtBranchOperator')
        ]
        task_ids_by_domain = [
            task_id for task_id in task_ids if task_id.startswith(domain_name) and not task_id.endswith('_increment')
        ]
        unittest.TestCase().assertSetEqual(
            raw_dbt_nodes_by_domain_schedule[domain_name][schedule],
            set(task_ids_by_domain),
            msg=f'{dag.dag_id}, {domain_name}',
        )


def test_check_if_all_dbt_models_are_in_dbt_af_graph(manifest, profiles, project_profile_name, config, get_etl_name):
    check_if_all_dbt_models_are_in_dbt_af_graph(manifest, profiles, project_profile_name, config, get_etl_name)



import datetime

from attrs import define, field
from croniter import croniter, croniter_range


@define(order=False)
class CronExpression:
    raw_cron_expression: str = field(validator=lambda _, __, val: croniter.is_valid(val))

    def __eq__(self, other):
        if not isinstance(other, (CronExpression, str)):
            return NotImplemented
        if isinstance(other, str):
            return self.raw_cron_expression == other
        return self.raw_cron_expression == other.raw_cron_expression

    def distance_between_two_runs(self) -> datetime.timedelta:
        """
        Calculate the time difference between two consecutive runs of the cron expression.
        """
        now = datetime.datetime.now()
        cron_iter = croniter(self.raw_cron_expression, now)
        next_run = cron_iter.get_next(datetime.datetime)
        after_next_run = cron_iter.get_next(datetime.datetime)
        return after_next_run - next_run

    def embeddings_number(self, other: 'CronExpression | None', is_upstream_bigger: bool) -> int:
        """
        Calculate how many times this cron expression can be embedded into another one over a given period.
        """
        if other is None:
            return 0
        if not isinstance(other, CronExpression):
            raise ValueError(f'Comparison must be with another CronExpression, got {other}')
        if is_upstream_bigger:
            return 0

        now = datetime.datetime.now()
        self_cron_iter = croniter(self.raw_cron_expression, now)
        start_interval = self_cron_iter.get_next(datetime.datetime)
        end_interval = self_cron_iter.get_next(datetime.datetime)

        dts = list(
            croniter_range(
                start_interval,
                end_interval,
                other.raw_cron_expression,
                ret_type=datetime.datetime,
                expand_from_start_time=True,
            )
        )
        if dts and dts[-1] == end_interval:
            # both ends are inclusive, so we need to remove the last element if it's equal to the end_interval
            dts.pop()

        return len(dts)

    def _split_cron_expression(self) -> list[str]:
        return self.raw_cron_expression.split()

    @property
    def minutes(self) -> int:
        minutes_part = self._split_cron_expression()[0]
        if minutes_part == '*':
            return 0
        return int(minutes_part)

    @property
    def hours(self) -> int:
        hours_part = self._split_cron_expression()[1]
        if hours_part == '*':
            return 0
        return int(hours_part)

    @property
    def days(self) -> int:
        days_part = self._split_cron_expression()[2]
        if days_part == '*':
            return 1
        return int(days_part)

    @property
    def months(self) -> int:
        months_part = self._split_cron_expression()[3]
        if months_part == '*':
            return 1
        return int(months_part)

    @property
    def weekdays(self) -> int:
        weekdays_part = self._split_cron_expression()[4]
        if weekdays_part == '*':
            return 0
        return int(weekdays_part)



import sys
from typing import Any, Literal

try:
    from pydantic.v1 import BaseModel, Field, root_validator
except ModuleNotFoundError:
    from pydantic import BaseModel, Field, root_validator


class Target(BaseModel):
    target_type: str = Field(alias='type')
    target_schema: str = Field(alias='schema')
    threads: int | str = Field(default=1)

    # here could be any database-specific connection details

    @root_validator
    def validate_target(cls, values: dict[str, Any]) -> dict[str, Any]:
        for key, value in values.items():
            if key in cls.__fields__:
                continue
            values[key] = value
        return values

    class Config:
        extra = 'allow'


class KubernetesToleration(BaseModel):
    key: str
    operator: str


class KubernetesTarget(Target):
    target_type: Literal['kubernetes'] = Field(alias='type')
    node_pool_selector_name: str
    node_pool: str
    image_name: str
    pod_cpu_guarantee: str
    pod_memory_guarantee: str
    tolerations: list[KubernetesToleration]


class VenvTarget(Target):
    target_type: Literal['venv'] = Field(alias='type')
    system_site_packages: bool
    requirements: list[str] = Field(default_factory=list)
    python_version: str = Field(default_factory=lambda: '.'.join(map(str, sys.version_info[:2])))
    pip_install_options: list[str] = Field(default_factory=list)
    index_urls: list[str] = Field(default_factory=list)
    inherit_env: bool = False


class Profile(BaseModel):
    target: str
    outputs: dict[str, VenvTarget | KubernetesTarget | Target]


class Profiles(BaseModel):
    profiles_config: dict[str, Any] = Field(alias='config', default_factory=dict)

    @root_validator
    def validate_profiles(cls, values: dict[str, Any]) -> dict[str, Any]:
        for key, value in values.items():
            if key in cls.__fields__:
                continue
            values[key] = Profile.parse_obj(value)
        return values

    class Config:
        extra = 'allow'

    def __getitem__(self, item):
        return getattr(self, item)



from datetime import datetime
from functools import partial
from typing import Callable

from croniter import croniter, croniter_range

from dbt_af.common.scheduling import BaseScheduleTag
from dbt_af.parser.dbt_node_model import WaitPolicy


class _TaskScheduleMapping:
    def __init__(self, wait_policy: WaitPolicy):
        self.wait_policy = wait_policy

        self._mapping = {}

    def is_registered(self, upstream_schedule_tag: BaseScheduleTag, downstream_schedule_tag: BaseScheduleTag) -> bool:
        return (upstream_schedule_tag.name, downstream_schedule_tag.name) in self._mapping

    def add(
        self,
        upstream_schedule_tag: BaseScheduleTag,
        downstream_schedule_tag: BaseScheduleTag,
        fn: Callable | list[Callable],
    ):
        if not isinstance(fn, list):
            fn = [fn]
        self._mapping[(upstream_schedule_tag.name, downstream_schedule_tag.name)] = fn
        return self

    @staticmethod
    def _update_upstream_cron_args(fn: partial, upstream: BaseScheduleTag, downstream: BaseScheduleTag):
        if fn is None:
            return

        fn.keywords.update({'self_schedule': downstream, 'upstream_schedule': upstream})

    def get(self, key: tuple[BaseScheduleTag, BaseScheduleTag], default=None) -> list[Callable | None]:
        """
        If the result function is None (or list of None),
        it means that there is no need to find a specific upstream task, wait for the last one.
        """
        if not isinstance(default, list):
            default = [default]
        stream_names = (key[0].name, key[1].name)
        fns = self._mapping.get(stream_names, default)
        for fn in fns:
            self._update_upstream_cron_args(fn, upstream=key[0], downstream=key[1])

        return fns


def calculate_task_to_wait_execution_date(
    execution_date: datetime,
    self_schedule: BaseScheduleTag,
    upstream_schedule: BaseScheduleTag,
    num_iter: int | None = None,
):
    """
    this function calculates the correct nearest execution date for the upstream task to wait for.

    :param num_iter: number of iterations to go back in time
        by default it's None, which means that the function will return the last possible execution date.
        this parameter is used for the 'all' wait policy to iterate over all possible execution dates
    """
    self_cron = self_schedule.cron_expression()
    upstream_cron = upstream_schedule.cron_expression()
    interval_stop_dttm: datetime = croniter(self_cron.raw_cron_expression, execution_date).get_next(datetime)

    if self_schedule < upstream_schedule:
        cron_iter = croniter(upstream_cron.raw_cron_expression, interval_stop_dttm)
        cron_iter.get_prev()
        return cron_iter.get_prev(datetime)
    if self_schedule == upstream_schedule:
        if self_schedule.timeshift == upstream_schedule.timeshift:
            return execution_date
        cron_iter = croniter(upstream_cron.raw_cron_expression, execution_date)
        return cron_iter.get_prev(datetime)

    all_dts = list(
        croniter_range(execution_date, interval_stop_dttm, upstream_cron.raw_cron_expression, ret_type=datetime)
    )

    if all_dts and all_dts[-1] == interval_stop_dttm:
        all_dts.pop()

    if num_iter is None:
        return all_dts[-1]

    return all_dts[num_iter]


GLOBAL_TASK_SCHEDULE_MAPPINGS = {policy: _TaskScheduleMapping(policy) for policy in WaitPolicy}



from .backfill_dag_components import BackfillDagModel, BackfillDagSnapshot  # noqa
from .dag_components import DagComponent, DagModel, DagSnapshot, LargeTest, MediumTests  # noqa
from .dbt_af_builder import DbtAfGraph, DomainDagsRegistry  # noqa
from .task_dependencies import DagDelayedDependencyRegistry, RegistryDomainDependencies  # noqa
from .dbt_model_path_graph_builder import DbtModelPathGraph  # noqa
from .domain_dag import BackfillDomainDag, DomainDag  # noqa

__all__ = [
    'BackfillDagModel',
    'BackfillDagSnapshot',
    'DagComponent',
    'DagModel',
    'DagSnapshot',
    'LargeTest',
    'MediumTests',
    'DbtAfGraph',
    'DomainDagsRegistry',
    'DagDelayedDependencyRegistry',
    'RegistryDomainDependencies',
    'DbtModelPathGraph',
    'BackfillDomainDag',
    'DomainDag',
]



import logging
from typing import Callable, Dict, Tuple

from dbt_af.conf import Config


def prepare_mcd_callbacks(config: Config) -> Tuple[Dict[str, Callable], Dict[str, Callable]]:
    if config.mcd is None:
        logging.warning('No airflow variable for Monte Carlo Data settings')
        return {}, {}

    if not config.mcd.callbacks_enabled:
        logging.info('Monte Carlo Data callbacks are disabled')
        return {}, {}

    try:
        from airflow_mcd.callbacks import mcd_callbacks
    except ModuleNotFoundError:
        logging.error('Package airflow-mcd for MonteCarloData callbacks is not installed')
        return {}, {}

    mcd_dag_callbacks = mcd_callbacks.dag_callbacks if not config.is_dev else {}
    mcd_task_callbacks = mcd_callbacks.task_callbacks if not config.is_dev else {}

    return mcd_dag_callbacks, mcd_task_callbacks



from typing import TYPE_CHECKING, Any, Callable, List, Optional

from airflow.operators.python import BranchPythonOperator
from airflow.utils.context import Context

from dbt_af.parser.dbt_node_model import DbtNodeConfig

if TYPE_CHECKING:
    from airflow.models import DAG
    from airflow.utils.task_group import TaskGroup

DOWNSTREAM_TASK_IDS_KWARG = 'downstream_task_ids'


class DbtBranchOperator(BranchPythonOperator):
    def __init__(
        self, task_id: str, task_group: 'Optional[TaskGroup]', python_callable: Callable, dag: 'DAG', **kwargs
    ):
        task_id = f'{task_id}_branch' if not task_id.endswith('_branch') else task_id
        super().__init__(task_id=task_id, task_group=task_group, python_callable=python_callable, dag=dag, **kwargs)

    def execute(self, context: Context) -> Any:
        downstream_task_ids = [t.task_id for t in self.downstream_list]
        self.op_kwargs[DOWNSTREAM_TASK_IDS_KWARG] = downstream_task_ids
        return super().execute(context)


def create_decision_path_function(node_config: DbtNodeConfig, node_name: str) -> Callable:
    def decide_which_path(**kwargs) -> List[str]:
        is_enable = True
        if node_config.enable_from_dttm:
            if str(kwargs['data_interval_end']) < node_config.enable_from_dttm:
                is_enable = False
        is_disable = False
        if node_config.disable_from_dttm:
            if str(kwargs['data_interval_start']) > node_config.disable_from_dttm:
                is_disable = True
        if is_enable and not is_disable:
            downstream = [f'{node_name}__group.{node_name}']
            if DOWNSTREAM_TASK_IDS_KWARG in kwargs:
                downstream += kwargs[DOWNSTREAM_TASK_IDS_KWARG]
            return downstream

        return []

    return decide_which_path



import logging

from dbt_af.conf import Config


def prepare_custom_af_callbacks(config: Config) -> tuple[dict[str, callable], dict[str, callable]]:
    if config.af_callbacks is None:
        logging.warning('No callback were passed')
        return {}, {}

    airflow_task_callbacks = {
        'on_success_callback': config.af_callbacks.task_on_success_callback,
        'on_failure_callback': config.af_callbacks.task_on_failure_callback,
        'on_retry_callback': config.af_callbacks.task_on_retry_callback,
        'on_execute_callback': config.af_callbacks.task_on_execute_callback,
    }

    airflow_dag_callbacks = {
        'on_failure_callback': config.af_callbacks.dag_on_failure_callback,
        'on_success_callback': config.af_callbacks.dag_on_success_callback,
        'sla_miss_callback': config.af_callbacks.dag_sla_miss_callback,
    }

    return airflow_dag_callbacks, airflow_task_callbacks



import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tableauserverclient.models.tableau_auth import Credentials

    from dbt_af.conf import Config


def get_tableau_auth_object(config: 'Config') -> 'Credentials':
    import tableauserverclient as tsc

    if not config.tableau:
        raise ValueError('No tableau configuration found in dbt_af config. Please specify tableau credentials.')

    if config.tableau.username and config.tableau.password:
        logging.info('Using tableau username/password authentication')
        return tsc.TableauAuth(
            username=config.tableau.username,
            password=config.tableau.password,
            site_id=config.tableau.site_id,
            user_id_to_impersonate=config.tableau.user_id_to_impersonate,
        )

    if config.tableau.token_name and config.tableau.pat:
        logging.info('Using tableau personal access token authentication')
        return tsc.PersonalAccessTokenAuth(
            token_name=config.tableau.token_name,
            personal_access_token=config.tableau.pat,
            site_id=config.tableau.site_id,
        )

    raise ValueError(
        'No valid tableau authentication method found in dbt_af config. Please specify tableau credentials.'
    )



from collections.abc import Iterable
from typing import Optional

from dbt_af.conf import Config
from dbt_af.integrations.af_callbacks import prepare_custom_af_callbacks
from dbt_af.integrations.mcd import prepare_mcd_callbacks


def merge_dicts_by_key(*dicts: dict) -> dict:
    merged_dict = dict()

    for d in dicts:
        for key, value in d.items():
            merged_dict.setdefault(key, [])
            if isinstance(value, Iterable):
                merged_dict[key].extend(list(value))
            else:
                merged_dict[key].append(value)
    return merged_dict


def collect_af_custom_callbacks(
    config: Config,
) -> tuple[dict[str, list[Optional[callable]]], dict[str, list[Optional[callable]]]]:
    mcd_dag_callbacks, mcd_task_callbacks = prepare_mcd_callbacks(config)
    airflow_dag_callbacks, airflow_task_callbacks = prepare_custom_af_callbacks(config)

    dag_callbacks = merge_dicts_by_key(mcd_dag_callbacks, airflow_dag_callbacks)
    task_callbacks = merge_dicts_by_key(mcd_task_callbacks, airflow_task_callbacks)

    return dag_callbacks, task_callbacks



import datetime as dt
import enum
from collections import defaultdict
from pathlib import Path
from typing import Any, DefaultDict, Dict, List, Literal, Optional, Union

import pendulum

try:
    import pydantic.v1 as pydantic
except ModuleNotFoundError:
    import pydantic

from dbt_af.common.constants import DOMAIN_DAG_START_DATE_FMT
from dbt_af.common.scheduling import BaseScheduleTag, EScheduleTag
from dbt_af.common.utils import TestTag
from dbt_af.conf import DbtDefaultTargetsConfig
from dbt_af.parser.dbt_profiles import KubernetesTarget, Profile, Target, VenvTarget


class DbtModelMaintenanceType(enum.Enum):
    PERSIST_DOCS = 'persist_docs'
    OPTIMIZE_TABLES = 'optimize_table'
    VACUUM_TABLE = 'vacuum_table'
    DEDUPLICATE_TABLE = 'deduplicate_table'
    SET_TTL_ON_TABLE = 'set_ttl_on_table'


class WaitPolicy(enum.Enum):
    last = 'last'
    all = 'all'


class DependencyConfig(pydantic.BaseModel):
    skip: bool = pydantic.Field(default=False)
    wait_policy: WaitPolicy = WaitPolicy.last


class TTLConfig(pydantic.BaseModel):
    key: str = pydantic.Field(description='Table timestamp-like column name')
    expiration_timeout: int = pydantic.Field(description='Table expiration timeout in days')
    additional_predicate: Optional[str] = pydantic.Field(
        description='Additional predicate for filtering expired data',
        default='',
    )
    force_predicate: Optional[str] = pydantic.Field(
        description='Force predicate for filtering expired data. It will override base predicate',
        default='',
    )

    @pydantic.root_validator(pre=True)
    def validate_ttl_config(cls, values: dict[str, Any]) -> dict[str, Any]:
        # if force_predicate is set, then key and expiration_timeout are not required
        if values.get('force_predicate'):
            values['key'] = values.get('key', '')
            values['expiration_timeout'] = values.get('expiration_timeout', 0)

        return values


class DbtAFMaintenanceConfig(pydantic.BaseModel):
    ttl: Optional[TTLConfig] = pydantic.Field(default=None)
    persist_docs: Optional[bool] = pydantic.Field(default=False)
    optimize_table: Optional[bool] = pydantic.Field(default=False)
    vacuum_table: Optional[bool] = pydantic.Field(default=False)
    deduplicate_table: Optional[bool] = pydantic.Field(default=False)

    def get_required_maintenance_types(self) -> list[DbtModelMaintenanceType]:
        maintenance_types = []
        if self.ttl is not None:
            maintenance_types.append(DbtModelMaintenanceType.SET_TTL_ON_TABLE)
        if self.persist_docs:
            maintenance_types.append(DbtModelMaintenanceType.PERSIST_DOCS)
        if self.optimize_table:
            maintenance_types.append(DbtModelMaintenanceType.OPTIMIZE_TABLES)
        if self.vacuum_table:
            maintenance_types.append(DbtModelMaintenanceType.VACUUM_TABLE)
        if self.deduplicate_table:
            maintenance_types.append(DbtModelMaintenanceType.DEDUPLICATE_TABLE)

        return maintenance_types


class TableauRefreshResourceType(str, enum.Enum):
    workbook = 'workbook'
    datasource = 'datasource'


class TableauRefreshTaskConfig(pydantic.BaseModel):
    resource_name: str
    project_name: str
    resource_type: TableauRefreshResourceType


class DbtNodeConfig(pydantic.BaseModel):
    enabled: bool
    alias: Optional[str]
    node_schema: Optional[str] = pydantic.Field(..., alias='schema')
    database: Optional[str]
    tags: List[str]
    meta: Dict[str, Any]
    group: Optional[str]
    materialized: Optional[str]
    incremental_strategy: Optional[str]
    persist_docs: Optional[Dict[str, Any]]
    quoting: Optional[Dict[str, Any]]
    column_types: Optional[Dict[str, Any]]
    full_refresh: Optional[bool]
    unique_key: Optional[Union[str, List[str]]]
    on_schema_change: Optional[str]
    grants: Optional[Dict]
    packages: Optional[List]
    docs: Optional[Dict[str, Any]]
    contract: Optional[Dict[str, Any]]
    file_format: Optional[str]
    tblproperties: Optional[Dict[str, Any]]
    partition_by: Optional[Union[str, List[str]]]
    incremental_predicates: Optional[List[str]]
    post_hook: Optional[List[Dict[str, Any]]] = pydantic.Field(alias='post-hook', default_factory=list)
    pre_hook: Optional[List[Dict[str, Any]]] = pydantic.Field(alias='pre-hook', default_factory=list)

    schedule: Optional[BaseScheduleTag] = pydantic.Field(default_factory=EScheduleTag.daily)
    schedule_shift: int = pydantic.Field(default=0)
    schedule_shift_unit: Literal['minute', 'hour', 'day'] = pydantic.Field(default='minute')

    dependencies: Optional[DefaultDict[str, DependencyConfig]] = pydantic.Field(
        default_factory=lambda: defaultdict(DependencyConfig)
    )

    env: dict[str, str] = pydantic.Field(default_factory=dict)

    py_cluster: Optional[str]
    sql_cluster: Optional[str]
    daily_sql_cluster: Optional[str]
    bf_cluster: Optional[str]

    enable_from_dttm: Optional[str] = pydantic.Field(default='')
    disable_from_dttm: Optional[str] = pydantic.Field(default='')

    airflow_parallelism: int = pydantic.Field(default=1)
    domain_start_date: Optional[str] = pydantic.Field(default='')

    dbt_target: Optional[str] = pydantic.Field(default='')

    maintenance: DbtAFMaintenanceConfig = pydantic.Field(default_factory=DbtAFMaintenanceConfig)

    tableau_refresh_tasks: Optional[List[TableauRefreshTaskConfig]] = pydantic.Field(default_factory=list)

    class Config:
        arbitrary_types_allowed = True

    @pydantic.root_validator(pre=True)
    def validate_values(cls, values):
        if values['materialized'] not in ('test', 'seed'):
            if not all(
                [
                    values['py_cluster'],
                    values['sql_cluster'],
                    values['daily_sql_cluster'],
                    values['bf_cluster'],
                ]
            ):
                raise ValueError('py_cluster, sql_cluster and daily_sql_cluster must be set')

        # build schedule with shift if any
        timeshift = None
        if (
            'schedule_shift' in values
            and values['schedule_shift'] > 0
            and 'schedule_shift_unit' in values
            and values['schedule_shift_unit']
            in (
                'minute',
                'hour',
                'day',
                'week',
            )
        ):
            timeshift = dt.timedelta(**{f'{values["schedule_shift_unit"]}s': values['schedule_shift']})
        if values.get('schedule') not in [item().name for item in EScheduleTag]:
            values['schedule'] = EScheduleTag.daily(timeshift=timeshift)
        else:
            values['schedule'] = EScheduleTag[values['schedule'].lstrip('@')](timeshift=timeshift)

        return values

    @pydantic.root_validator(pre=True)
    def validate_airflow_parallelism(cls, values):
        if 'airflow_parallelism' in values and isinstance(values['airflow_parallelism'], str):
            values['airflow_parallelism'] = int(values['airflow_parallelism'])
        return values

    @pydantic.validator('domain_start_date', pre=True)
    def validate_domain_start_date(cls, v):
        if v == '':
            return v
        pendulum.from_format(v, DOMAIN_DAG_START_DATE_FMT)
        return v


class DbtNodeColumn(pydantic.BaseModel):
    name: str
    description: Optional[str]
    meta: Optional[Dict[str, Any]]
    data_type: Optional[str]
    constraints: Optional[List]
    quote: Optional[bool]
    tags: Optional[List[str]]


class DbtNode(pydantic.BaseModel):
    database: str
    node_schema: str = pydantic.Field(..., alias='schema')
    name: str
    resource_type: str
    package_name: str
    path: str
    original_file_path: str
    unique_id: str
    fqn: List[str]
    alias: Optional[str]
    checksum: Dict[str, str]
    config: DbtNodeConfig
    tags: List[str]
    description: Optional[str]
    columns: Optional[Dict[str, DbtNodeColumn]]
    meta: Optional[Dict[str, Any]]
    group: Optional[str]
    docs: Optional[Dict[str, Any]]
    patch_path: Optional[str]
    build_path: Optional[str]
    deferred: Optional[bool]
    unrendered_config: Optional[Dict[str, Any]]
    created_at: float
    relation_name: Optional[str]
    raw_code: str
    language: Optional[str]
    refs: Optional[List[Dict[str, Any]]]
    sources: Optional[List[Any]] = pydantic.Field(default_factory=list)
    metrics: Optional[List[Dict[str, Any]]] = pydantic.Field(default_factory=list)
    node_depends_on: Optional[Dict[str, Any]] = pydantic.Field(alias='depends_on', default_factory=dict)
    compiled_path: Optional[str]
    contract: Optional[Dict[str, Any]] = pydantic.Field(default_factory=dict)
    access: Optional[str]
    constraints: Optional[List[Dict[str, Any]]] = pydantic.Field(default_factory=list)
    version: Optional[str]
    latest_version: Optional[str]

    target_details: Optional[Target | KubernetesTarget | VenvTarget] = pydantic.Field(default=None)

    def __hash__(self):
        return hash(self.unique_id)

    def __eq__(self, other):
        if isinstance(other, DbtNode):
            return self.unique_id == other.unique_id
        if isinstance(other, str):
            return self.unique_id == other
        raise TypeError(f'Cannot compare {self.__class__.__name__} with {other.__class__.__name__}')

    @property
    def domain(self):
        """Gets domain name from a dbt resource fqn

        The function assumes that the layout of project looks something like:
            `[<project>, models, <domain>, <layer>, <model>, <model>]`

        This results in the fqn that looks like:
            `[<project>, <domain>, <layer>, <model>, <model>]` for models
            `[<project>, dbt, models, <domain>, <layer>, <model>, <model>]` for tests
        """
        if self.is_test():
            return self.fqn[3]
        return self.fqn[1]

    @property
    def resource_name(self):
        """Gets resource name from dbt's node_id

        test.dtt.not_null_task_w_test__ods__one_id.1417bf4cb1
        model.dtt.task_w_test__ods__one

        """
        if self.unique_id.startswith('test.'):
            return self.unique_id.split('.', maxsplit=2)[2].rsplit('.', maxsplit=1)[0]

        return self.unique_id.split('.', maxsplit=2)[2]

    @property
    def depends_on(self):
        return [
            dep for dep in self.node_depends_on['nodes'] if dep.startswith(('test.', 'model.', 'snapshot.', 'seed.'))
        ]

    @property
    def depends_on_sources(self):
        return [dep for dep in self.node_depends_on['nodes'] if dep.startswith('source.')]

    @property
    def materialized(self):
        return self.config.materialized if self.resource_type == 'model' else ''

    @property
    def test_type(self):
        # TODO refactor this
        if not self.is_test():
            return None

        test_tags = {f.value for f in TestTag} & set(self.tags)
        if len(test_tags) > 1:
            raise ValueError(f'Several test tags for {self.unique_id} are not allowed')
        if not test_tags:
            return TestTag.small.value

        return test_tags.pop()

    @property
    def model_type(self):
        if self.is_py_model():
            return 'py'
        if self.is_sql_model():
            return 'sql'
        return None

    def target_environment(self, default_dbt_targets: DbtDefaultTargetsConfig) -> str:
        if self.config.dbt_target:
            return self.config.dbt_target

        if self.is_test():
            return default_dbt_targets.default_for_tests_target

        if len(self.config.pre_hook) == 0 and self.model_type == 'sql':
            if self.config.schedule in (EScheduleTag.daily(), EScheduleTag.weekly()):
                return self.config.daily_sql_cluster

            return self.config.sql_cluster

        return self.config.py_cluster

    @property
    def original_file_path_dirname(self) -> str:
        return str(Path(self.original_file_path).parent)

    @property
    def original_file_path_without_extension(self):
        return str(Path(self.original_file_path).with_suffix(''))

    def is_model(self) -> bool:
        return self.resource_type == 'model'

    def is_view(self) -> bool:
        return self.materialized == 'view'

    def is_snapshot(self) -> bool:
        return self.resource_type == 'snapshot'

    def is_test(self) -> bool:
        return self.resource_type == 'test'

    def is_small_test(self) -> bool:
        return self.resource_type == 'test' and self.test_type == TestTag.small.value

    def is_medium_test(self) -> bool:
        return self.resource_type == 'test' and self.test_type == TestTag.medium.value

    def is_large_test(self) -> bool:
        return self.resource_type == 'test' and self.test_type == TestTag.large.value

    def is_seed(self) -> bool:
        return self.resource_type == 'seed'

    def is_sql_model(self) -> bool:
        return self.path.endswith('.sql')

    def is_py_model(self) -> bool:
        return self.path.endswith('.py')

    def is_at_etl_service(self, etl_service_name) -> bool:
        return any(part == etl_service_name for part in self.original_file_path.split('/'))

    def has_dt_partition(self) -> bool:
        # TODO: refactor this
        whitelisted_dt_partitions = {'timestamp'}

        if self.config.partition_by is None:
            return False
        if isinstance(self.config.partition_by, str):
            return '_dt' in self.config.partition_by or self.config.partition_by in whitelisted_dt_partitions
        if isinstance(self.config.partition_by, list):
            return '_dt' in ''.join(self.config.partition_by) or not whitelisted_dt_partitions.isdisjoint(
                self.config.partition_by
            )

        return False

    def get_airflow_parallelism(self) -> int:
        if self.materialized != 'incremental' or not self.has_dt_partition():
            return 1

        return self.config.airflow_parallelism

    def get_required_maintenance_types(self) -> list[DbtModelMaintenanceType]:
        return self.config.maintenance.get_required_maintenance_types()

    def set_target_details(self, profile: Profile, default_dbt_targets: DbtDefaultTargetsConfig):
        self.target_details = profile.outputs[self.target_environment(default_dbt_targets=default_dbt_targets)]



def test_af_wait_name_is_less_250_chars(manifest):
    """
    Check that wait name is less than 250 chars. 5 chars are reserved for suffix with pattern __{dep_number}
    where 0 <= dep_number <= 999
    """

    for node in manifest['nodes'].values():
        if node['resource_type'] == 'model':
            for dep in node['depends_on']['nodes']:
                if dep.startswith('model'):
                    dep_name = manifest['nodes'][dep]['name']
                    dep_safe_name = dep_name.replace('.', '__')
                    dep_domain_name = dep_name.split('.')[0]
                    wait_name = f'{dep_domain_name}__scheduletag__dependencies__group.wait__{dep_safe_name}'
                    assert len(wait_name) <= 245
        elif node['resource_type'] == 'test':
            for dep in node['depends_on']['nodes']:
                if dep.startswith('model'):
                    model_name_safe_name = '.'.join(dep.split('.')[2:]).replace('.', '__')
                    wait_name = f"{model_name_safe_name}__group.{node['name']}"
                    assert len(wait_name) <= 250



from pathlib import Path

import pytest
import typer

cli = typer.Typer()


@cli.command()
def run_tests(
    manifest_path: str = typer.Option(exists=True),
    profiles_path: str = typer.Option(exists=True),
    dbt_project_path: str = typer.Option(exists=True),
    target: str = typer.Option(exists=True),
    etl_name: str = typer.Option(None),
):
    args = [
        Path(__file__).parent.as_posix(),
        '-q',
        '-s',
        '-vv',
        '--manifest_path',
        manifest_path,
        '--profiles_path',
        profiles_path,
        '--dbt_project_path',
        dbt_project_path,
        '--target',
        target,
    ]
    if etl_name:
        args.extend(['--etl_name', etl_name])
    raise SystemExit(pytest.main(args))


if __name__ == '__main__':
    cli()



import json
import os
import shutil
from pathlib import Path

import typer
import yaml

from dbt_af.builder.dbt_model_path_graph_builder import DbtModelPathGraph

cli = typer.Typer()


@cli.command()
def generate_mini_dbt_projects(manifest_path: str, dbt_project_path: str, result_path: str):
    generate_mini_dbt_template(dbt_project_path, result_path, 'mini_dbt_template')

    with open(manifest_path) as fin:
        manifest = json.load(fin)

    graph = DbtModelPathGraph.from_manifest(manifest)

    for node in graph.dbt_nodes:
        # copy template

        shutil.copytree(
            Path(result_path, 'mini_dbt_template'),
            Path(result_path, node.resource_name),
        )

        # create list of dirs we should copy for model
        list_of_directories = []

        # model path
        list_of_directories.append(graph.node_to_path[node.unique_id])
        # model deps pathes
        list_of_directories.extend(graph.full_path_to_path[graph.node_to_path[node.unique_id]])

        for path in list_of_directories:
            from_path = Path(dbt_project_path, path)
            to_path = Path(result_path, node.resource_name, path)

            # create destination dir
            to_path.mkdir(parents=True, exist_ok=True)

            # copy all sql
            for file in from_path.glob('*.sql'):
                shutil.copy(file, to_path)

            # copy all py
            for file in from_path.glob('*.py'):
                shutil.copy(file, to_path)

            # copy all yml
            for file in from_path.glob('*.yml'):
                shutil.copy(file, to_path)

    return graph


def generate_mini_dbt_template(dbt_project_path, result_path, mini_dbt_node_template):
    with open(Path(dbt_project_path, 'dbt_project.yml')) as fin:
        dbt_project = yaml.safe_load(fin)

    Path(result_path, mini_dbt_node_template).mkdir(parents=True, exist_ok=True)
    # copy core files
    shutil.copyfile(
        Path(dbt_project_path, 'dbt_project.yml'),
        Path(result_path, mini_dbt_node_template, 'dbt_project.yml'),
    )
    shutil.copyfile(
        Path(dbt_project_path, 'packages.yml'),
        Path(result_path, mini_dbt_node_template, 'packages.yml'),
    )

    # copy any md file into one md per folder
    for path in dbt_project['model-paths']:
        from_path = Path(dbt_project_path, path)
        to_path = Path(result_path, mini_dbt_node_template, path)

        to_path.mkdir(parents=True, exist_ok=True)
        md_pathlist = from_path.glob('**/*.md')
        with open(Path(to_path, 'doc.md'), 'w') as outfile:
            for md_path in md_pathlist:
                with open(md_path) as infile:
                    for line in infile:
                        outfile.write(line)

        for yml_file in from_path.glob('**/*.yml'):
            file_to_path = Path(str(yml_file).replace(str(from_path), str(to_path)))
            with open(yml_file) as fin:
                some_yml = yaml.safe_load(fin)
                if 'groups' in some_yml:
                    os.makedirs(os.path.dirname(file_to_path), exist_ok=True)
                    shutil.copy(yml_file, file_to_path)

    # all macro
    for path in dbt_project['macro-paths']:
        from_path = Path(dbt_project_path, path)
        to_path = Path(result_path, mini_dbt_node_template, path)
        if os.path.exists(to_path):
            shutil.rmtree(to_path)
        shutil.copytree(from_path, to_path)


if __name__ == '__main__':
    cli()



import json
import logging
import shutil
from datetime import timedelta
from tempfile import TemporaryDirectory
from typing import Dict, Optional

try:
    import pydantic.v1 as pydantic
except ModuleNotFoundError:
    import pydantic

from airflow.operators.bash import BashOperator
from airflow.utils.context import Context

from dbt_af.common.constants import DBT_COMPILE_POOL
from dbt_af.common.scheduling import BaseScheduleTag, EScheduleTag
from dbt_af.common.utils import find_latest_log_file, init_environment
from dbt_af.conf import Config, RetryPolicy


def get_delay_by_schedule(schedule_tag):
    if schedule_tag is not None and schedule_tag == EScheduleTag.hourly():
        return {'execution_timeout': timedelta(minutes=60)}
    return {'execution_timeout': timedelta(hours=6)}


class DbtBaseOperator(BashOperator):
    @property
    def cli_command(self) -> str:
        raise NotImplementedError()

    def _patch_path_to_dbt_bash(self, **kwargs) -> str:
        return 'PATH_TO_DBT=$DBT_PROJECT_DIR && '

    def generate_bash(self, **kwargs) -> str:
        return (
            self._patch_path_to_dbt_bash(**kwargs)
            + ' cd $PATH_TO_DBT  && {dbt_executable_path} {debug} {cli} '
            '--profiles-dir $DBT_PROFILES_DIR '
            '--project-dir $PATH_TO_DBT '
            '--target {target_environment}'.format(dbt_executable_path=self.dbt_af_config.dbt_executable_path, **kwargs)
        )

    def __init__(
        self,
        dbt_af_config: Config,
        schedule_tag: BaseScheduleTag,
        target_environment: str,
        max_active_tis_per_dag: int = 1,  # only one parallel tasks in the universe
        pool: Optional[str] = None,
        retry_policy: Optional[RetryPolicy] = None,
        env: dict[str, str] | None = None,
        **kwargs,
    ) -> None:
        self.dbt_af_config = dbt_af_config

        self.debug = '--debug' if self.dbt_af_config.debug_mode_enabled else ''
        self.cli = self.cli_command

        self.target_environment = target_environment or dbt_af_config.dbt_default_targets.default_target
        assert self.target_environment, 'Target environment must be specified'

        kwargs.update(get_delay_by_schedule(schedule_tag))
        af_pool = pool or f'dbt_{self.target_environment}' if dbt_af_config.use_dbt_target_specific_pools else None

        retry_policy = (
            retry_policy.as_dict()
            if retry_policy is not None
            else dbt_af_config.retries_config.default_retry_policy.as_dict()
        )
        super().__init__(
            max_active_tis_per_dag=max_active_tis_per_dag,
            env=init_environment(self.dbt_af_config) | (env or {}),
            pool=af_pool,
            append_env=True,
            bash_command=self.generate_bash(**self.__dict__),
            **retry_policy,
            **kwargs,
        )

    def execute(self, context: Context):
        with TemporaryDirectory(dir=self.dbt_af_config.dbt_project.dbt_target_path) as tmp_target_path:
            # copy manifest.json to tmp target path to isolate it
            shutil.copy(
                self.dbt_af_config.dbt_project.dbt_project_path / 'target/manifest.json',
                f'{tmp_target_path}/manifest.json',
            )

            self.bash_command += f' --target-path {tmp_target_path}'
            if self.dbt_af_config.is_dev:
                # there is no dry-run mode in dbt, so we use -h flag just for empty dbt run
                self.bash_command += ' -h'

            super().execute(context)

            if self.dbt_af_config.mcd and self.dbt_af_config.mcd.artifacts_export_enabled:
                from dbt_af.integrations.mcd import send_dbt_artifacts_to_montecarlo

                latest_log_file = find_latest_log_file(context, self.dbt_af_config.dbt_project.dbt_log_path)
                try:
                    send_dbt_artifacts_to_montecarlo(
                        target_path=tmp_target_path,
                        log_path=latest_log_file,
                        model_name=context['task'].task_id,
                        metastore_name=self.dbt_af_config.mcd.metastore_name,
                        project_name=self.dbt_af_config.dbt_project.dbt_project_name,
                    )
                except Exception as e:
                    logging.warning(f'Failed to send dbt artifacts to MonteCarlo: {e}')
                    if self.dbt_af_config.mcd.success_required:
                        raise e


class DbtConstOperator(DbtBaseOperator):
    def __init__(self, pool: str = DBT_COMPILE_POOL, **kwargs) -> None:
        task_id = f'dbt_{self.cli_command.replace(" ", "_")}'
        super().__init__(pool=pool, task_id=task_id, **kwargs)


class DbtCompile(DbtConstOperator):
    @property
    def cli_command(self) -> str:
        return 'compile'


class DbtParse(DbtConstOperator):
    @property
    def cli_command(self) -> str:
        return 'parse'


class DbtDeps(DbtConstOperator):
    @property
    def cli_command(self) -> str:
        return 'deps'


class DbtClean(DbtConstOperator):
    @property
    def cli_command(self) -> str:
        return 'clean'


class DbtDocsGenerate(DbtConstOperator):
    @property
    def cli_command(self) -> str:
        return 'docs generate'


class DbtModelVars(pydantic.BaseModel):
    start_dttm: str
    end_dttm: str
    overlap: bool = False
    extra: Dict = pydantic.Field(exclude=True)

    _raw_keys_to_drop = {'start_dttm', 'end_dttm'}
    _dbt_keys_to_patch = {'dbt_start_dttm', 'dbt_end_dttm'}

    @pydantic.root_validator(pre=True)
    def build_extra(cls, values: Dict):
        required_fields = [field.alias for field in cls.__fields__.values() if field.alias != 'extra']

        extra = {}
        for field_name in list(values):
            if field_name in cls._raw_keys_to_drop:
                continue
            if field_name in cls._dbt_keys_to_patch:
                values[field_name.strip('dbt_')] = values[field_name]
            if field_name not in required_fields:
                extra[field_name] = values.pop(field_name)
                if isinstance(extra[field_name], (list, tuple)):
                    # brackets [] are not supported for array values in Databricks
                    # HACK: if it's empty collection, we pass empty string. It will work ONLY with string fields, but
                    # it's impossible to fetch correct type of the field from Databricks, and it's impossible to
                    # pass just empty brackets
                    extra[field_name] = (
                        f"({','.join(map(json.dumps, extra[field_name]))})" if len(extra[field_name]) > 0 else '("")'
                    )

        values['extra'] = extra
        return values

    def dict(self, **kwargs):
        """
        Updates the dict representation of the model with the extra fields.
        """
        return dict(**super().dict(**kwargs), **self.extra)


class DbtIntervalActionOperator(DbtBaseOperator):
    overlap = False

    def execute(self, context):
        updated_context = self._time_delta_logic(context)

        self.log.debug(
            'Context params:\n%s',
            '\n'.join(f'{k}={v}' for k, v in updated_context['params'].items()),
        )

        self.bash_command += (
            f" --vars '{json.dumps(DbtModelVars(**updated_context['params'], overlap=self.overlap).dict())}'"
        )

        super().execute(updated_context)

    @staticmethod
    def _time_delta_logic(context):
        """
        Here we parse the time interval and set the dbt_start_dttm and dbt_end_dttm variables
        If we get start_dttm and end_dttm in the context, we use them (in this case we get them from user-defined
        parameters)
        """
        # user defined parameters
        if 'start_dttm' in context['params'] and 'end_dttm' in context['params']:
            if context['params']['start_dttm'] != context['params']['end_dttm']:
                context['params']['dbt_start_dttm'] = context['params']['start_dttm']
                context['params']['dbt_end_dttm'] = context['params']['end_dttm']
                return context

        # if data_interval_start is equal to data_interval_end, we run dbt for the whole day starting
        # from data_interval_start.date()
        if context['data_interval_start'] == context['data_interval_end']:
            ds = context['data_interval_start'].replace(hour=0, minute=0, second=0, microsecond=0)
            de = ds + timedelta(days=1)
            context['params']['dbt_start_dttm'] = ds.isoformat()
            context['params']['dbt_end_dttm'] = de.isoformat()
            return context

        # take start_dttm == data_interval_start and end_dttm == data_interval_end (most common use case)
        context['params']['dbt_start_dttm'] = context['data_interval_start'].isoformat()
        context['params']['dbt_end_dttm'] = context['data_interval_end'].isoformat()

        return context


class DbtBaseActionOperator(DbtIntervalActionOperator):
    def generate_bash(self, **kwargs):
        base_bash = super().generate_bash(**kwargs)
        return base_bash + ' --select {model_name}'.format(**kwargs)

    def __init__(
        self,
        model_name: str,
        model_type: str = '',
        schedule_tag: Optional[BaseScheduleTag] = None,
        overlap: bool = False,
        **kwargs,
    ) -> None:
        self.model_name = f'{model_name}.{model_type}' if model_type else model_name
        self.model_name_wo_type = model_name
        self.overlap = overlap

        super().__init__(schedule_tag=schedule_tag, **kwargs)

    def _patch_path_to_dbt_bash(self, **kwargs) -> str:
        return (
            'if [[ "$DBT_ENABLE_MINI_DBT" = "true" ]]; '
            'then PATH_TO_DBT="$DBT_MINI_DBT_DIR/{model_name_wo_type}"; '
            'else PATH_TO_DBT="$DBT_PROJECT_DIR"; fi && '
        ).format(**kwargs)



import datetime
import re
from abc import ABC, abstractmethod
from enum import Enum
from functools import cache

from dbt_af.common.cron import CronExpression


class BaseScheduleTag(ABC):
    default_timeshift = datetime.timedelta()

    def __init__(self, timeshift: datetime.timedelta | None = None):
        if timeshift is not None:
            if self.default_cron_expression is None:
                raise ValueError('Cannot set timeshift for manual schedule tag')
            if timeshift >= self.default_cron_expression.distance_between_two_runs():
                raise ValueError(f'Cannot shift schedule tag "{self.base_name}" by {timeshift}')
        self.timeshift = timeshift or self.default_timeshift

    @property
    @abstractmethod
    def base_name(self) -> str:
        raise NotImplementedError()

    @property
    def name(self) -> str:
        if self.timeshift == self.default_timeshift or self.timeshift is None:
            return self.base_name

        full_days_shift, full_hours_shift, rest_minutes_shift = self.split_timeshift()
        shifts = [
            ('day', full_days_shift),
            ('hour', full_hours_shift),
            ('minute', rest_minutes_shift),
        ]
        shift_parts = [f'{value}_{unit}s' for unit, value in shifts if value > 0]
        if shift_parts:
            return f'{self.base_name}_shift_{"_".join(shift_parts)}'

        return self.base_name

    @property
    @abstractmethod
    def level(self) -> int:
        """
        Is used for comparison of schedule tags.
        Must be unique for each schedule tag.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def default_cron_expression(self) -> CronExpression | None:
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def _af_repr_impl(full_days_shift: int, full_hours_shift: int, rest_minutes_shift: int) -> str | None:
        raise NotImplementedError()

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __lt__(self, other):
        if not isinstance(other, BaseScheduleTag):
            return NotImplemented
        if self.level == other.level:
            return self.timeshift < other.timeshift
        return self.level < other.level

    def __eq__(self, other: 'str | BaseScheduleTag'):
        if isinstance(other, str):
            return self.base_name == other
        if isinstance(other, BaseScheduleTag):
            return self.base_name == other.base_name
        raise TypeError(f'Cannot compare {self} with {other}')

    @property
    def safe_name(self):
        return re.sub(r'^@', 'dbt_', self.name)

    def split_timeshift(self) -> tuple[int, int, int]:
        full_days_shift = self.timeshift.days
        full_hours_shift = self.timeshift.seconds // 3600
        rest_minutes_shift = (self.timeshift.seconds % 3600) // 60

        return full_days_shift, full_hours_shift, rest_minutes_shift

    @cache
    def af_repr(self) -> str | None:
        """
        Returns airflow-like schedule representation.
        Could be cron expression or None if there's no schedule.
        """
        full_days_shift, full_hours_shift, rest_minutes_shift = self.split_timeshift()
        return self._af_repr_impl(full_days_shift, full_hours_shift, rest_minutes_shift)

    def cron_expression(self) -> CronExpression | None:
        af_schedule = self.af_repr()
        if af_schedule is None:
            return None
        return CronExpression(af_schedule)


class _MonthlyScheduleTag(BaseScheduleTag):
    base_name = '@monthly'
    default_cron_expression = CronExpression('0 0 1 * *')
    level = 5

    def _af_repr_impl(self, full_days_shift: int, full_hours_shift: int, rest_minutes_shift: int):
        """
        base cron expression: 0 0 1 * * (or @monthly)
        """

        if full_days_shift > 31:
            raise ValueError(f'Monthly schedule tag supports only shifts between 1 and 31 days, got {self.timeshift}')
        if full_hours_shift > 23:
            raise ValueError(f'Monthly schedule tag supports only shifts between 0 and 23 hours, got {self.timeshift}')
        if rest_minutes_shift > 59:
            raise ValueError(
                f'Monthly schedule tag supports only shifts between 0 and 59 minutes, got {self.timeshift}'
            )
        if full_days_shift == 0:
            full_days_shift = '1'

        return f'{rest_minutes_shift} {full_hours_shift} {full_days_shift} * *'


class _WeeklyScheduleTag(BaseScheduleTag):
    base_name = '@weekly'
    default_cron_expression = CronExpression('0 0 * * 0')
    level = 4

    def _af_repr_impl(self, full_days_shift: int, full_hours_shift: int, rest_minutes_shift: int):
        """
        base cron expression: 0 0 * * 0 (or @weekly)
        """

        if full_days_shift > 6:
            raise ValueError(f'Weekly schedule tag supports only shifts between 0 and 6 days, got {self.timeshift}')

        return f'{rest_minutes_shift} {full_hours_shift} * * {full_days_shift}'


class _DailyScheduleTag(BaseScheduleTag):
    base_name = '@daily'
    default_cron_expression = CronExpression('0 0 * * *')
    level = 3

    def _af_repr_impl(self, full_days_shift: int, full_hours_shift: int, rest_minutes_shift: int):
        """
        base cron expression: 0 0 * * * (or @daily)
        """

        if full_days_shift:
            raise ValueError(f'Daily schedule tag supports only shifts between 0 and 23 hours, got {self.timeshift}')

        return f'{rest_minutes_shift} {full_hours_shift} * * *'


class _HourlyScheduleTag(BaseScheduleTag):
    base_name = '@hourly'
    default_cron_expression = CronExpression('0 * * * *')
    level = 2

    def _af_repr_impl(self, full_days_shift: int, full_hours_shift: int, rest_minutes_shift: int):
        """
        base cron expression: 0 * * * * (or @hourly)
        """

        if full_days_shift or full_hours_shift:
            raise ValueError(f'Hourly schedule tag supports only shifts between 0 and 59 minutes, got {self.timeshift}')

        return f'{rest_minutes_shift} * * * *'


class _Every15MinutesScheduleTag(BaseScheduleTag):
    base_name = '@every15minutes'
    default_cron_expression = CronExpression('*/15 * * * *')
    level = 1

    def _af_repr_impl(self, full_days_shift: int, full_hours_shift: int, rest_minutes_shift: int):
        """
        base cron expression: */15 * * * * (or @15minutes)
        """

        if full_days_shift or full_hours_shift or rest_minutes_shift > 59:
            raise ValueError(
                f'Every 15 minutes schedule tag supports only shifts between 0 and 59 minutes, got {self.timeshift}'
            )
        return f'{rest_minutes_shift}-59/15 * * * *'


class _ManualScheduleTag(BaseScheduleTag):
    base_name = '@manual'
    default_cron_expression = None
    level = 0

    def __init__(self, timeshift: datetime.timedelta | None = None):
        super().__init__(timeshift)
        self.timeshift = None

    @staticmethod
    def _af_repr_impl(full_days_shift: int, full_hours_shift: int, rest_minutes_shift: int):
        return None

    def af_repr(self):
        # there's no schedule for manual DAGs
        return None


class EScheduleTag(Enum):
    """
    Schedule tags for dbt models.
    Only these tags are supported.

    Enum's key must be equal to the EScheduleTag name.
    """

    monthly = _MonthlyScheduleTag
    weekly = _WeeklyScheduleTag
    daily = _DailyScheduleTag
    hourly = _HourlyScheduleTag
    manual = _ManualScheduleTag
    every15minutes = _Every15MinutesScheduleTag

    @property
    def name(self):
        return self.value.name

    @property
    def base_name(self):
        return self.value.base_name

    @property
    def level(self):
        return self.value.level

    @property
    def default_cron_expression(self):
        return self.value.default_cron_expression

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return self.value.name

    def __lt__(self, other):
        if not isinstance(other, EScheduleTag):
            return NotImplemented
        return self.value.level < other.value.level

    def __eq__(self, other):
        if isinstance(other, EScheduleTag):
            return self.value.name == other.value.name
        if isinstance(other, str):
            return self.value.name == other
        raise TypeError(f'Cannot compare {self} with {other}')

    def __call__(self, *args, **kwargs):
        return self.value(*args, **kwargs)



from typing import TYPE_CHECKING, Optional

from airflow import Dataset

from dbt_af.common.constants import DBT_MODEL_DAG_PARAM
from dbt_af.conf import Config
from dbt_af.operators.base import DbtBaseActionOperator

if TYPE_CHECKING:
    from airflow.utils.context import Context


class DbtBaseDatasetOperator(DbtBaseActionOperator):
    def __init__(self, model_name: Optional[str], is_dataset_enable=False, model_type: str = 'sql', **kwargs) -> None:
        if model_name:
            # exactly one model
            super().__init__(
                model_name=model_name,
                model_type=model_type,
                outlets=[Dataset(model_name)] if is_dataset_enable else [],
                **kwargs,
            )
        else:
            super().__init__(model_name=DBT_MODEL_DAG_PARAM, **kwargs)

    def execute(self, context: 'Context'):
        if 'params' in context:
            if DBT_MODEL_DAG_PARAM in context['params'] and self.model_name == DBT_MODEL_DAG_PARAM:
                self.bash_command = self.bash_command.replace(
                    DBT_MODEL_DAG_PARAM, context['params'][DBT_MODEL_DAG_PARAM]
                )

        super().execute(context)

    def _patch_path_to_dbt_bash(self, **kwargs):
        if self.model_name_wo_type == DBT_MODEL_DAG_PARAM:
            return 'PATH_TO_DBT=$DBT_PROJECT_DIR && '
        return super()._patch_path_to_dbt_bash(**kwargs)


class DbtRun(DbtBaseDatasetOperator):
    @property
    def cli_command(self) -> str:
        return 'run'

    def __init__(self, dbt_af_config: 'Config', **kwargs) -> None:
        super().__init__(
            dbt_af_config=dbt_af_config,
            retry_policy=dbt_af_config.retries_config.dbt_run_retry_policy,
            **kwargs,
        )


class DbtSeed(DbtBaseDatasetOperator):
    @property
    def cli_command(self) -> str:
        return 'seed'

    def __init__(self, dbt_af_config: 'Config', **kwargs) -> None:
        super().__init__(
            dbt_af_config=dbt_af_config,
            retry_policy=dbt_af_config.retries_config.dbt_seed_retry_policy,
            **kwargs,
        )


class DbtSnapshot(DbtBaseDatasetOperator):
    @property
    def cli_command(self) -> str:
        return 'snapshot'

    def __init__(self, dbt_af_config: 'Config', **kwargs) -> None:
        super().__init__(
            dbt_af_config=dbt_af_config,
            retry_policy=dbt_af_config.retries_config.dbt_snapshot_retry_policy,
            **kwargs,
        )


class DbtTest(DbtBaseActionOperator):
    @property
    def cli_command(self) -> str:
        return 'test'

    def __init__(self, dbt_af_config: 'Config', **kwargs) -> None:
        super().__init__(
            dbt_af_config=dbt_af_config,
            max_active_tis_per_dag=None,
            target_environment=dbt_af_config.dbt_default_targets.default_for_tests_target,
            retry_policy=dbt_af_config.retries_config.dbt_test_retry_policy,
            overlap=True,
            **kwargs,
        )
